/**********************************************************************
 * $Id: cpl_string.h,v 1.6 2014/12/16 10:27:51 cvsbw Exp $
 *
 * Name:     cpl_string.h
 * Project:  CPL - Common Portability Library
 * Purpose:  String and StringList functions.
 * Author:   Daniel Morissette, dmorissette@mapgears.com
 *
 **********************************************************************
 * Copyright (c) 1998, Daniel Morissette
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#ifndef _CPL_STRING_H_INCLUDED
#define _CPL_STRING_H_INCLUDED

#include "cpl_vsi.h"
#include "cpl_error.h"
#include "cpl_conv.h"

/**
 * \file cpl_string.h
 *
 * Various convenience functions for working with strings and string lists. 
 *
 * A StringList is just an array of strings with the last pointer being
 * NULL.  An empty StringList may be either a NULL pointer, or a pointer to
 * a pointer memory location with a NULL value.
 *
 * A common convention for StringLists is to use them to store name/value
 * lists.  In this case the contents are treated like a dictionary of
 * name/value pairs.  The actual data is formatted with each string having
 * the format "<name>:<value>" (though "=" is also an acceptable separator). 
 * A number of the functions in the file operate on name/value style
 * string lists (such as CSLSetNameValue(), and CSLFetchNameValue()).
 *
 * To some extent the CPLStringList C++ class can be used to abstract
 * managing string lists a bit but still be able to return them from C
 * functions.
 *
 */

CPL_C_START

char CPL_DLL **CSLAddString(char **papszStrList, const char *pszNewString) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CSLCount(char **papszStrList);
const char CPL_DLL *CSLGetField( char **, int );
void CPL_DLL CPL_STDCALL CSLDestroy(char **papszStrList);
char CPL_DLL **CSLDuplicate(char **papszStrList) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLMerge( char **papszOrig, char **papszOverride ) CPL_WARN_UNUSED_RESULT;

char CPL_DLL **CSLTokenizeString(const char *pszString ) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLTokenizeStringComplex(const char *pszString,
                                   const char *pszDelimiter,
                                   int bHonourStrings, int bAllowEmptyTokens ) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLTokenizeString2( const char *pszString, 
                                   const char *pszDelimeter, 
                                   int nCSLTFlags ) CPL_WARN_UNUSED_RESULT;

#define CSLT_HONOURSTRINGS      0x0001
#define CSLT_ALLOWEMPTYTOKENS   0x0002
#define CSLT_PRESERVEQUOTES     0x0004
#define CSLT_PRESERVEESCAPES    0x0008
#define CSLT_STRIPLEADSPACES    0x0010
#define CSLT_STRIPENDSPACES     0x0020

int CPL_DLL CSLPrint(char **papszStrList, FILE *fpOut);
char CPL_DLL **CSLLoad(const char *pszFname) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLLoad2(const char *pszFname, int nMaxLines, int nMaxCols, char** papszOptions) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CSLSave(char **papszStrList, const char *pszFname);

char CPL_DLL **CSLInsertStrings(char **papszStrList, int nInsertAtLineNo, 
                         char **papszNewLines) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLInsertString(char **papszStrList, int nInsertAtLineNo, 
                               const char *pszNewLine) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **CSLRemoveStrings(char **papszStrList, int nFirstLineToDelete,
                         int nNumToRemove, char ***ppapszRetStrings) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CSLFindString( char **, const char * );
int CPL_DLL CSLPartialFindString( char **papszHaystack, 
	const char * pszNeedle );
int CPL_DLL CSLFindName(char **papszStrList, const char *pszName);
int CPL_DLL CSLTestBoolean( const char *pszValue );
int CPL_DLL CSLFetchBoolean( char **papszStrList, const char *pszKey, 
                             int bDefault );

const char CPL_DLL *CPLSPrintf(const char *fmt, ...) CPL_PRINT_FUNC_FORMAT(1, 2);
char CPL_DLL **CSLAppendPrintf(char **papszStrList, const char *fmt, ...) CPL_PRINT_FUNC_FORMAT(2, 3) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CPLVASPrintf(char **buf, const char *fmt, va_list args );

const char CPL_DLL *
      CPLParseNameValue(const char *pszNameValue, char **ppszKey );
const char CPL_DLL *
      CSLFetchNameValue(char **papszStrList, const char *pszName);
const char CPL_DLL *
      CSLFetchNameValueDef(char **papszStrList, const char *pszName,
                           const char *pszDefault );
char CPL_DLL **
      CSLFetchNameValueMultiple(char **papszStrList, const char *pszName);
char CPL_DLL **
      CSLAddNameValue(char **papszStrList, 
                      const char *pszName, const char *pszValue) CPL_WARN_UNUSED_RESULT;
char CPL_DLL **
      CSLSetNameValue(char **papszStrList, 
                      const char *pszName, const char *pszValue) CPL_WARN_UNUSED_RESULT;
void CPL_DLL CSLSetNameValueSeparator( char ** papszStrList, 
                                       const char *pszSeparator );

#define CPLES_BackslashQuotable 0
#define CPLES_XML               1
#define CPLES_URL               2
#define CPLES_SQL               3
#define CPLES_CSV               4
#define CPLES_XML_BUT_QUOTES    5

char CPL_DLL *CPLEscapeString( const char *pszString, int nLength, 
                               int nScheme ) CPL_WARN_UNUSED_RESULT;
char CPL_DLL *CPLUnescapeString( const char *pszString, int *pnLength,
                                 int nScheme ) CPL_WARN_UNUSED_RESULT;

char CPL_DLL *CPLBinaryToHex( int nBytes, const GByte *pabyData ) CPL_WARN_UNUSED_RESULT;
GByte CPL_DLL *CPLHexToBinary( const char *pszHex, int *pnBytes ) CPL_WARN_UNUSED_RESULT;

char CPL_DLL *CPLBase64Encode( int nBytes, const GByte *pabyData ) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CPLBase64DecodeInPlace(GByte* pszBase64);

typedef enum
{
    CPL_VALUE_STRING,
    CPL_VALUE_REAL,
    CPL_VALUE_INTEGER
} CPLValueType;

CPLValueType CPL_DLL CPLGetValueType(const char* pszValue);

size_t CPL_DLL CPLStrlcpy(char* pszDest, const char* pszSrc, size_t nDestSize);
size_t CPL_DLL CPLStrlcat(char* pszDest, const char* pszSrc, size_t nDestSize);
size_t CPL_DLL CPLStrnlen (const char *pszStr, size_t nMaxLen);

/* -------------------------------------------------------------------- */
/*      RFC 23 character set conversion/recoding API (cpl_recode.cpp).  */
/* -------------------------------------------------------------------- */
#define CPL_ENC_LOCALE     ""
#define CPL_ENC_UTF8       "UTF-8"
#define CPL_ENC_UTF16      "UTF-16"
#define CPL_ENC_UCS2       "UCS-2"
#define CPL_ENC_UCS4       "UCS-4"
#define CPL_ENC_ASCII      "ASCII"
#define CPL_ENC_ISO8859_1  "ISO-8859-1"

int CPL_DLL  CPLEncodingCharSize( const char *pszEncoding );
void CPL_DLL  CPLClearRecodeWarningFlags();
char CPL_DLL *CPLRecode( const char *pszSource, 
                         const char *pszSrcEncoding, 
                         const char *pszDstEncoding ) CPL_WARN_UNUSED_RESULT;
char CPL_DLL *CPLRecodeFromWChar( const wchar_t *pwszSource, 
                                  const char *pszSrcEncoding, 
                                  const char *pszDstEncoding ) CPL_WARN_UNUSED_RESULT;
wchar_t CPL_DLL *CPLRecodeToWChar( const char *pszSource,
                                   const char *pszSrcEncoding, 
                                   const char *pszDstEncoding ) CPL_WARN_UNUSED_RESULT;
int CPL_DLL CPLIsUTF8(const char* pabyData, int nLen);
char CPL_DLL *CPLForceToASCII(const char* pabyData, int nLen, char chReplacementChar) CPL_WARN_UNUSED_RESULT;

CPL_C_END

/************************************************************************/
/*                              CPLString                               */
/************************************************************************/

#if defined(__cplusplus) && !defined(CPL_SUPRESS_CPLUSPLUS)

#include <string>

/*
 * Simple trick to avoid "using" declaration in header for new compilers
 * but make it still working with old compilers which throw C2614 errors.
 *
 * Define MSVC_OLD_STUPID_BEHAVIOUR
 * for old compilers: VC++ 5 and 6 as well as eVC++ 3 and 4.
 */

/*
 * Detect old MSVC++ compiler <= 6.0
 * 1200 - VC++ 6.0
 * 1200-1202 - eVC++ 4.0
 */
#if defined(_MSC_VER) 
# if (_MSC_VER <= 1202) 
#  define MSVC_OLD_STUPID_BEHAVIOUR 
# endif
#endif

/* Avoid C2614 errors */
#ifdef MSVC_OLD_STUPID_BEHAVIOUR
    using std::string;
# define gdal_std_string string
#else
# define gdal_std_string std::string
#endif 

/* Remove annoying warnings in Microsoft eVC++ and Microsoft Visual C++ */
#if defined(WIN32CE)
#  pragma warning(disable:4251 4275 4786)
#endif

//! Convenient string class based on std::string.
#if defined _MSC_VER && _MSC_VER >= 1600

class CPL_DLL CPLString
{
public:
    class Impl;
private:
    Impl *pimpl;
public:
          Impl& impl();
    const Impl& impl() const;
    CPLString( const Impl & );

    typedef char value_type;
    typedef char* pointer;
    typedef char& reference;
    typedef const char& const_reference;
    typedef std::size_t size_type;
    typedef std::ptrdiff_t difference_type;
    
    static const size_type &npos;
  
    CPLString(const CPLString& s);
    CPLString(const CPLString& s, size_type pos, size_type n = npos);
    CPLString(const char* s, size_type n);
    CPLString(size_type n, char c);
    ~CPLString();

    CPLString& operator=(const CPLString&);
    CPLString& operator=(const char* s);
    CPLString& operator=(char c);
    CPLString& assign(const CPLString&);
    CPLString& assign(const CPLString& s, size_type pos, size_type n);
    CPLString& assign(const char* s, size_type n);
    CPLString& assign(const char* s);
    CPLString& assign(size_type n, char c);

          char& at( size_type pos );
		const char& at( size_type pos ) const;
    const char* data() const;
    const char* c_str() const;

    bool empty() const;
    size_type size() const;
    size_type length() const;
    size_type max_size() const;
    void reserve(size_t);
    size_type capacity() const;

    void clear();
    CPLString& insert(size_type pos, const CPLString& s);
    CPLString& insert(size_type pos, const CPLString& s, size_type pos1, size_type n);
    CPLString& insert(size_type pos, const char* s);
    CPLString& insert(size_type pos, const char* s, size_type n);
    CPLString& insert(size_type pos, size_type n, char c);
    CPLString& erase(size_type pos = 0, size_type n = npos);
    void push_back(char c);
    CPLString& append(const CPLString& s);
    CPLString& append(const CPLString& s, size_type pos, size_type n);
    CPLString& append(const char* s);
    CPLString& append(const char* s, size_type n);
    CPLString& append(size_type n, char c);
    CPLString& operator+=(const CPLString& s);
    CPLString& operator+=(const char* s);
    CPLString& operator+=(char c);
    int compare(const CPLString& s) const;
    int compare(size_type pos, size_type n, const CPLString& s) const;
    int compare(size_type pos, size_type n, const CPLString& s, size_type pos1, size_type n1) const;
    int compare(const char* s) const;
    int compare(size_type pos, size_type n, const char* s, size_type len = npos) const;
    CPLString& replace(size_type pos, size_type n, const CPLString& s);
    CPLString& replace(size_type pos, size_type n, const CPLString& s, size_type pos1, size_type n1);
    CPLString& replace(size_type pos, size_type n, const char* s, size_type n1);
    CPLString& replace(size_type pos, size_type n, const char* s);
    CPLString& replace(size_type pos, size_type n, size_type n1, char c);
    CPLString substr(size_type pos = 0, size_type n = npos) const;
    size_type copy(char* buf, size_type n, size_type pos = 0) const;
    void resize(size_type n, char c = char());
    void swap(CPLString&);

    size_type find(const CPLString& s, size_type pos = 0) const;
    size_type find(const char* s, size_type pos, size_type n) const;
    size_type find(const char* s, size_type pos = 0) const;
    size_type find(char c, size_type pos = 0) const;
    size_type rfind(const CPLString& s, size_type pos = npos) const;
    size_type rfind(const char* s, size_type pos, size_type n) const;
    size_type rfind(const char* s, size_type pos = npos) const;
    size_type rfind(char c, size_type pos = npos) const;
    size_type find_first_of(const CPLString& s, size_type pos = 0) const;
    size_type find_first_of(const char* s, size_type pos, size_type n) const;
    size_type find_first_of(const char* s, size_type pos = 0) const;
    size_type find_first_of(char c, size_type pos = 0) const;
    size_type find_first_not_of(const CPLString& s, size_type pos = 0) const;
    size_type find_first_not_of(const char* s, size_type pos, size_type n) const;
    size_type find_first_not_of(const char* s, size_type pos = 0) const;
    size_type find_first_not_of(char c, size_type pos = 0) const;
    size_type find_last_of(const CPLString& s, size_type pos = npos) const;
    size_type find_last_of(const char* s, size_type pos, size_type n) const;
    size_type find_last_of(const char* s, size_type pos = npos) const;
    size_type find_last_of(char c, size_type pos = npos) const;
    size_type find_last_not_of(const CPLString& s, size_type pos = npos) const;
    size_type find_last_not_of(const char* s, size_type pos, size_type n) const;
    size_type find_last_not_of(const char* s, size_type pos = npos) const;
    size_type find_last_not_of(char c, size_type pos = npos) const;

    operator ::std::string&();
    operator const ::std::string&() const;

#else

class CPL_DLL CPLString : public gdal_std_string
{
public:
  typedef std::string::size_type size_type;

#endif

    CPLString(void);
    CPLString( const std::string &oStr );
    CPLString( const char *pszStr );

    operator const char* (void) const;// { return c_str(); }

          char& operator[](size_type i);
    const char& operator[](size_type i) const;

          char& operator[](int i);
    const char& operator[](int i) const;

    void Clear();

    // NULL safe assign and free.
    void Seize(char *pszValue);

    /* There seems to be a bug in the way the compiler count indices... Should be CPL_PRINT_FUNC_FORMAT (1, 2) */
    CPLString &Printf( const char *pszFormat, ... ) CPL_PRINT_FUNC_FORMAT (2, 3);
    CPLString &vPrintf( const char *pszFormat, va_list args );
    CPLString &FormatC( double dfValue, const char *pszFormat = NULL );
    CPLString &Trim();
    CPLString &Recode( const char *pszSrcEncoding, const char *pszDstEncoding );

    /* case insensitive find alternates */
    size_t    ifind( const std::string & str, size_t pos = 0 ) const;
    size_t    ifind( const char * s, size_t pos = 0 ) const;
    CPLString &toupper( void );
    CPLString &tolower( void );
};

#if defined _MSC_VER && _MSC_VER >= 1600

// wk 20141215: here comes weird usage of operator+, necessary e.g. for ogrcsvdatasource.cpp(140)
// The following overload does not need to be defined in the original version of CPLString, 
// because CPLString::operator const char *() is defined in CPLString itself,
// while operator+( std::string, char ) is defined only in its base class.
// Here, we define both operators in CPLString, and thus, the compiler cannot guess which one to select.
// Equivalent to: s.substr(i)
CPL_DLL CPLString operator+(const CPLString &s, int i);
CPL_DLL CPLString operator+(const CPLString &s, size_t i);

CPL_DLL CPLString operator+(const CPLString& s1, const CPLString& s2);
CPL_DLL CPLString operator+(const char* s1, const CPLString& s2);
CPL_DLL CPLString operator+(const CPLString& s1, const char* s2);
CPL_DLL CPLString operator+(char c, const CPLString& s2);
CPL_DLL CPLString operator+(const CPLString& s1, char c);
CPL_DLL bool operator==(const CPLString& s1, const CPLString& s2);
CPL_DLL bool operator==(const char* s1, const CPLString& s2);
CPL_DLL bool operator==(const CPLString& s1, const char* s2);
CPL_DLL bool operator!=(const CPLString& s1, const CPLString& s2);
CPL_DLL bool operator!=(const char* s1, const CPLString& s2);
CPL_DLL bool operator!=(const CPLString& s1, const char* s2);
CPL_DLL bool operator<(const CPLString& s1, const CPLString& s2);
CPL_DLL bool operator<(const char* s1, const CPLString& s2);
CPL_DLL bool operator<(const CPLString& s1, const char* s2);
CPL_DLL void swap(CPLString& s1, CPLString& s2);
CPL_DLL std::istream& operator>>(std::istream& is, CPLString& s);
CPL_DLL std::ostream& operator<<(std::ostream& os, CPLString& s);
CPL_DLL std::istream& getline(std::istream& is, CPLString& s, char delim);
CPL_DLL std::istream& getline(std::istream& is, CPLString& s);
    
#endif

/* -------------------------------------------------------------------- */
/*      URL processing functions, here since they depend on CPLString.  */
/* -------------------------------------------------------------------- */
CPLString CPL_DLL CPLURLGetValue(const char* pszURL, const char* pszKey);
CPLString CPL_DLL CPLURLAddKVP(const char* pszURL, const char* pszKey,
                               const char* pszValue);

/************************************************************************/
/*                            CPLStringList                             */
/************************************************************************/

//! String list class designed around our use of C "char**" string lists.
class CPL_DLL CPLStringList
{
    char **papszList;
    mutable int nCount;
    mutable int nAllocation;
    int    bOwnList;
    int    bIsSorted;

    void   Initialize();
    void   MakeOurOwnCopy();
    void   EnsureAllocation( int nMaxLength );
    int    FindSortedInsertionPoint( const char *pszLine );
    
  public:
    CPLStringList();
    CPLStringList( char **papszList, int bTakeOwnership=TRUE );
    CPLStringList( const CPLStringList& oOther );
    ~CPLStringList();

    CPLStringList &Clear();

    int    size() const { return Count(); }
    int    Count() const;

    CPLStringList &AddString( const char *pszNewString );
    CPLStringList &AddStringDirectly( char *pszNewString );

    CPLStringList &InsertString( int nInsertAtLineNo, const char *pszNewLine )
    { return InsertStringDirectly( nInsertAtLineNo, CPLStrdup(pszNewLine) ); }
    CPLStringList &InsertStringDirectly( int nInsertAtLineNo, char *pszNewLine);
    
//    CPLStringList &InsertStrings( int nInsertAtLineNo, char **papszNewLines );
//    CPLStringList &RemoveStrings( int nFirstLineToDelete, int nNumToRemove=1 );
    
    int    FindString( const char *pszTarget ) const
    { return CSLFindString( papszList, pszTarget ); }
    int    PartialFindString( const char *pszNeedle ) const
    { return CSLPartialFindString( papszList, pszNeedle ); }

    int    FindName( const char *pszName ) const;
    int    FetchBoolean( const char *pszKey, int bDefault ) const;
    const char *FetchNameValue( const char *pszKey ) const;
    const char *FetchNameValueDef( const char *pszKey, const char *pszDefault ) const;
    CPLStringList &AddNameValue( const char *pszKey, const char *pszValue );
    CPLStringList &SetNameValue( const char *pszKey, const char *pszValue );

    CPLStringList &Assign( char **papszList, int bTakeOwnership=TRUE );
    CPLStringList &operator=(char **papszListIn) { return Assign( papszListIn, TRUE ); }
    CPLStringList &operator=(const CPLStringList& oOther);

    char * operator[](int i);
    char * operator[](size_t i) { return (*this)[(int)i]; }
    const char * operator[](int i) const;
    const char * operator[](size_t i) const { return (*this)[(int)i]; }

    char **List() { return papszList; }
    char **StealList();

    CPLStringList &Sort();
    int    IsSorted() const { return bIsSorted; }

    operator char**(void) { return List(); }
};

#endif /* def __cplusplus && !CPL_SUPRESS_CPLUSPLUS */

#endif /* _CPL_STRING_H_INCLUDED */
