/******************************************************************************
 * $Id: cplstring.cpp,v 1.5 2014/12/16 10:27:51 cvsbw Exp $
 *
 * Project:  GDAL 
 * Purpose:  CPLString implementation.
 * Author:   Frank Warmerdam, warmerdam@pobox.com
 *
 ******************************************************************************
 * Copyright (c) 2005, Frank Warmerdam <warmerdam@pobox.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/

#include "cpl_string.h"
#include <string>

CPL_CVSID("$Id: cplstring.cpp,v 1.5 2014/12/16 10:27:51 cvsbw Exp $");

/*
 * The CPLString class is derived from std::string, so the vast majority 
 * of the implementation comes from that.  This module is just the extensions
 * we add. 
 */

#if defined _MSC_VER && _MSC_VER >= 1600

const CPLString::size_type& CPLString::npos = std::string::npos;

class CPLString::Impl : public std::string
{
  typedef std::string Base;
public:
  Impl() : Base() {}
  Impl( const std::string &oStr ) : Base( oStr ) {}
  Impl( const char *pszStr ) : Base( pszStr ) {}
  Impl( const Impl& s, size_type pos, size_type n ) : Base( s, pos, n ) {}
  Impl( const char* s, size_type n ) : Base( s, n ) {}
  Impl( size_type n, char c ) : Base( n, c ) {}

  Impl& operator=(const Impl &ref)
  {
    Base::operator=(ref);
    return *this;
  }
  Impl& operator=(const char* s)
  {
    Base::operator=(s);
    return *this;
  }
  Impl& operator=(char c)
  {
    Base::operator=(c);
    return *this;
  }

};

      CPLString::Impl& CPLString::impl()       { return *pimpl; }
const CPLString::Impl& CPLString::impl() const { return *pimpl; }

CPLString::CPLString( const Impl &ref ) : pimpl( new Impl( ref ) ) {}
CPLString::CPLString() : pimpl( new Impl() ) {}
CPLString::CPLString( const CPLString& s ) : pimpl( new Impl( s.impl() ) ) {}
CPLString::CPLString( const std::string &oStr ) : pimpl( new Impl( oStr ) ) {}
CPLString::CPLString( const char *pszStr ) : pimpl( new Impl( pszStr ) ) {}
CPLString::CPLString( const CPLString& s, size_type pos, size_type n ) : pimpl( new Impl( s.impl(), pos, n ) ) {}
CPLString::CPLString( const char* s, size_type n ) : pimpl( new Impl( s, n ) ) {}
CPLString::CPLString( size_type n, char c ) : pimpl( new Impl( n, c ) ) {}

CPLString::~CPLString() { delete pimpl; }

CPLString& CPLString::operator=(const CPLString &ref) { pimpl->operator=( ref.impl() ); return *this; }
CPLString& CPLString::operator=(const char* s) { pimpl->operator=(s); return *this; }
CPLString& CPLString::operator=(char c) { pimpl->operator=(c); return *this; }

CPLString& CPLString::assign( const CPLString &s ) { pimpl->assign( s.impl() ); return *this; }
CPLString& CPLString::assign( const CPLString &s, size_type pos, size_type n ) { pimpl->assign( s.impl(), pos, n ); return *this; }
CPLString& CPLString::assign(const char* s, size_type n) { pimpl->assign( s, n ); return *this; }
CPLString& CPLString::assign(const char* s) { pimpl->assign( s ); return *this; }
CPLString& CPLString::assign(size_type n, char c) { pimpl->assign( n, c ); return *this; }

      char& CPLString::at( size_type pos )       { return pimpl->at( pos ); }
const char& CPLString::at( size_type pos ) const { return pimpl->at( pos ); }
      char& CPLString::operator[](size_type i)       { return pimpl->operator[]( i ); }
const char& CPLString::operator[](size_type i) const { return pimpl->operator[]( i ); }

const char* CPLString::data() const { return pimpl->data(); }
const char* CPLString::c_str() const { return pimpl->c_str(); }

bool CPLString::empty() const { return pimpl->empty(); }
CPLString::size_type CPLString::size() const { return pimpl->size(); }
CPLString::size_type CPLString::length() const { return pimpl->length(); }
CPLString::size_type CPLString::max_size() const { return pimpl->max_size(); }
void CPLString::reserve(size_t s) { pimpl->reserve(s); }
CPLString::size_type CPLString::capacity() const { return pimpl->capacity(); }

void CPLString::clear() { pimpl->clear(); }
CPLString& CPLString::insert(size_type pos, const CPLString& s) { pimpl->insert( pos, s.impl() ); return *this; }
CPLString& CPLString::insert(size_type pos, const CPLString& s, size_type pos1, size_type n) { pimpl->insert( pos, s.impl(), pos1, n ); return *this; }
CPLString& CPLString::insert(size_type pos, const char* s) { pimpl->insert( pos, s ); return *this; }
CPLString& CPLString::insert(size_type pos, const char* s, size_type n) { pimpl->insert( pos, s, n ); return *this; }
CPLString& CPLString::insert(size_type pos, size_type n, char c) { pimpl->insert( pos, n, c ); return *this; }
CPLString& CPLString::erase(size_type pos, size_type n) { pimpl->erase( pos, n ); return *this; }
void CPLString::push_back(char c) { pimpl->push_back( c ); }
CPLString& CPLString::append(const CPLString& s) { pimpl->append( s.impl() ); return *this; }
CPLString& CPLString::append(const CPLString& s, size_type pos, size_type n) { pimpl->append( s.impl(), pos, n ); return *this; }
CPLString& CPLString::append(const char* s) { pimpl->append(s); return *this; }
CPLString& CPLString::append(const char* s, size_type n) { pimpl->append( s, n ); return *this; }
CPLString& CPLString::append(size_type n, char c) { pimpl->append( n, c ); return *this; }
CPLString& CPLString::operator+=( const CPLString& s ) { pimpl->operator+=( s.impl() ); return *this; }
CPLString& CPLString::operator+=( const char* s ) { pimpl->operator+=( s ); return *this; }
CPLString& CPLString::operator+=( char c ) { pimpl->operator+=( c ); return *this; }
int CPLString::compare(const CPLString& s) const { return pimpl->compare( s.impl() ); }
int CPLString::compare(size_type pos, size_type n, const CPLString& s) const { return pimpl->compare( pos, n, s.impl() ); }
int CPLString::compare(size_type pos, size_type n, const CPLString& s, size_type pos1, size_type n1) const { return pimpl->compare( pos, n, s.impl(), pos1, n1 ); }
int CPLString::compare(const char* s) const { return pimpl->compare( s ); }
int CPLString::compare(size_type pos, size_type n, const char* s, size_type len) const { return pimpl->compare( pos, n, s, len ); }
CPLString& CPLString::replace(size_type pos, size_type n, const CPLString& s) { pimpl->replace( pos, n, s.impl() ); return *this; }
CPLString& CPLString::replace(size_type pos, size_type n, const CPLString& s, size_type pos1, size_type n1) { pimpl->replace( pos, n, s.impl(), pos1, n1 ); return *this; }
CPLString& CPLString::replace(size_type pos, size_type n, const char* s, size_type n1) { pimpl->replace( pos, n, s, n1 ); return *this;  }
CPLString& CPLString::replace(size_type pos, size_type n, const char* s) { pimpl->replace( pos, n, s ); return *this; }
CPLString& CPLString::replace(size_type pos, size_type n, size_type n1, char c) { pimpl->replace( pos, n, n1, c ); return *this; }
CPLString CPLString::substr(size_type pos, size_type n) const { return pimpl->substr( pos, n ); }
CPLString::size_type CPLString::copy(char* buf, size_type n, size_type pos) const { return pimpl->copy( buf, n, pos ); }
void CPLString::resize(size_type n, char c) { pimpl->resize( n, c ); }
void CPLString::swap(CPLString &ref) { pimpl->swap( ref.impl() ); }

CPLString::size_type CPLString::find(const CPLString& s, size_type pos) const { return pimpl->find( s.impl(), pos ); }
CPLString::size_type CPLString::find(const char* s, size_type pos, size_type n) const { return pimpl->find( s, pos, n ); }
CPLString::size_type CPLString::find(const char* s, size_type pos) const { return pimpl->find( s, pos ); }
CPLString::size_type CPLString::find(char c, size_type pos) const { return pimpl->find( c, pos ); }
CPLString::size_type CPLString::rfind(const CPLString& s, size_type pos) const { return pimpl->rfind( s.impl(), pos ); }
CPLString::size_type CPLString::rfind(const char* s, size_type pos, size_type n) const { return pimpl->rfind( s, pos, n ); }
CPLString::size_type CPLString::rfind(const char* s, size_type pos) const { return pimpl->rfind( s, pos ); }
CPLString::size_type CPLString::rfind(char c, size_type pos) const { return pimpl->rfind( c, pos ); }
CPLString::size_type CPLString::find_first_of(const CPLString& s, size_type pos) const { return pimpl->find_first_of( s.impl(), pos ); }
CPLString::size_type CPLString::find_first_of(const char* s, size_type pos, size_type n) const { return pimpl->find_first_of( s, pos, n ); }
CPLString::size_type CPLString::find_first_of(const char* s, size_type pos) const { return pimpl->find_first_of( s, pos ); }
CPLString::size_type CPLString::find_first_of(char c, size_type pos) const { return pimpl->find_first_of( c, pos ); }
CPLString::size_type CPLString::find_first_not_of(const CPLString& s, size_type pos) const { return pimpl->find_first_not_of( s.impl(), pos ); }
CPLString::size_type CPLString::find_first_not_of(const char* s, size_type pos, size_type n) const { return pimpl->find_first_not_of( s, pos, n ); }
CPLString::size_type CPLString::find_first_not_of(const char* s, size_type pos) const { return pimpl->find_first_not_of( s, pos ); }
CPLString::size_type CPLString::find_first_not_of(char c, size_type pos) const { return pimpl->find_first_not_of( c, pos ); }
CPLString::size_type CPLString::find_last_of(const CPLString& s, size_type pos) const { return pimpl->find_last_of( s.impl(), pos ); }
CPLString::size_type CPLString::find_last_of(const char* s, size_type pos, size_type n) const { return pimpl->find_last_of( s, pos, n ); }
CPLString::size_type CPLString::find_last_of(const char* s, size_type pos) const { return pimpl->find_last_of( s, pos ); }
CPLString::size_type CPLString::find_last_of(char c, size_type pos) const { return pimpl->find_last_of( c, pos ); }
CPLString::size_type CPLString::find_last_not_of(const CPLString& s, size_type pos) const { return pimpl->find_last_not_of( s.impl(), pos ); }
CPLString::size_type CPLString::find_last_not_of(const char* s, size_type pos, size_type n) const { return pimpl->find_last_not_of( s, pos, n ); }
CPLString::size_type CPLString::find_last_not_of(const char* s, size_type pos) const { return pimpl->find_last_not_of( s, pos ); }
CPLString::size_type CPLString::find_last_not_of(char c, size_type pos) const { return pimpl->find_last_not_of( c, pos ); }

CPLString::operator       ::std::string&()       { return impl(); }
CPLString::operator const ::std::string&() const { return impl(); }

CPLString operator+(const CPLString &s, int i) { return s.c_str() + i; }
CPLString operator+(const CPLString &s, size_t i) { return s.c_str() + i; }

CPLString operator+(const CPLString& s1, const CPLString& s2) { return operator+( s1.impl(), s2.impl() ); }
CPLString operator+(const char* s1, const CPLString& s2) { return operator+( s1, s2.impl() ); }
CPLString operator+(const CPLString& s1, const char* s2) { return operator+( s1.impl(), s2 ); }
CPLString operator+(char c, const CPLString& s2) { return operator+( c, s2.impl() ); }
CPLString operator+(const CPLString& s1, char c) { return operator+( s1.impl(), c ); }
bool operator==(const CPLString& s1, const CPLString& s2) { return operator==( s1.impl(), s2.impl() ); }
bool operator==(const char* s1, const CPLString& s2) { return operator==( s1, s2.impl() ); }
bool operator==(const CPLString& s1, const char* s2) { return operator==( s1.impl(), s2 ); }
bool operator!=(const CPLString& s1, const CPLString& s2) { return operator!=( s1.impl(), s2.impl() ); }
bool operator!=(const char* s1, const CPLString& s2) { return operator!=( s1, s2.impl() ); }
bool operator!=(const CPLString& s1, const char* s2) { return operator!=( s1.impl(), s2 ); }
bool operator<(const CPLString& s1, const CPLString& s2) { return operator<( s1.impl(), s2.impl() ); }
bool operator<(const char* s1, const CPLString& s2) { return operator<( s1, s2.impl() ); }
bool operator<(const CPLString& s1, const char* s2) { return operator<( s1.impl(), s2 ); }
void swap(CPLString& s1, CPLString& s2) { swap( s1.impl(), s2.impl() ); }
std::istream& operator>>(std::istream& is, CPLString& s) { is >> s.impl(); return is; }
std::ostream&  operator<<(std::ostream& os, CPLString& s) { os << s.impl(); return os; }
std::istream& getline(std::istream& is, CPLString& s, char delim) { getline( is, s.impl(), delim ); return is; }
std::istream& getline(std::istream& is, CPLString& s) { getline( is, s.impl() ); return is; }

#else

char& CPLString::operator[](size_type i)
{
  return this->std::string::operator[](i);
}

const char& CPLString::operator[](size_type i) const
{
  return this->std::string::operator[](i);
}

#endif

CPLString::operator const char* () const { return c_str(); }

char& CPLString::operator[](int i)
{
  return this->operator[](static_cast<size_type>(i));
}

const char& CPLString::operator[](int i) const
{
  return this->operator[](static_cast<size_type>(i));
}

void CPLString::Clear() { resize(0); }

void CPLString::Seize(char *pszValue)
{
  if (pszValue == NULL )
    Clear();
  else
  {
    *this = pszValue;
    CPLFree(pszValue);
  }
}


/************************************************************************/
/*                               Printf()                               */
/************************************************************************/

CPLString &CPLString::Printf( const char *pszFormat, ... )

{
    va_list args;

    va_start( args, pszFormat );
    vPrintf( pszFormat, args );
    va_end( args );

    return *this;
}

/************************************************************************/
/*                              vPrintf()                               */
/************************************************************************/

CPLString &CPLString::vPrintf( const char *pszFormat, va_list args )

{
/* -------------------------------------------------------------------- */
/*      This implementation for platforms without vsnprintf() will      */
/*      just plain fail if the formatted contents are too large.        */
/* -------------------------------------------------------------------- */

#if !defined(HAVE_VSNPRINTF)
    char *pszBuffer = (char *) CPLMalloc(30000);
    if( vsprintf( pszBuffer, pszFormat, args) > 29998 )
    {
        CPLError( CE_Fatal, CPLE_AppDefined, 
                  "CPLString::vPrintf() ... buffer overrun." );
    }
    *this = pszBuffer;
    CPLFree( pszBuffer );

/* -------------------------------------------------------------------- */
/*      This should grow a big enough buffer to hold any formatted      */
/*      result.                                                         */
/* -------------------------------------------------------------------- */
#else
    char szModestBuffer[500];
    int nPR;
    va_list wrk_args;

#ifdef va_copy
    va_copy( wrk_args, args );
#else
    wrk_args = args;
#endif
    
    nPR = vsnprintf( szModestBuffer, sizeof(szModestBuffer), pszFormat, 
                     wrk_args );
    if( nPR == -1 || nPR >= (int) sizeof(szModestBuffer)-1 )
    {
        int nWorkBufferSize = 2000;
        char *pszWorkBuffer = (char *) CPLMalloc(nWorkBufferSize);

#ifdef va_copy
        va_end( wrk_args );
        va_copy( wrk_args, args );
#else
        wrk_args = args;
#endif
        while( (nPR=vsnprintf( pszWorkBuffer, nWorkBufferSize, pszFormat,wrk_args))
               >= nWorkBufferSize-1 
               || nPR == -1 )
        {
            nWorkBufferSize *= 4;
            pszWorkBuffer = (char *) CPLRealloc(pszWorkBuffer, 
                                                nWorkBufferSize );
#ifdef va_copy
            va_end( wrk_args );
            va_copy( wrk_args, args );
#else
            wrk_args = args;
#endif
        }
        *this = pszWorkBuffer;
        CPLFree( pszWorkBuffer );
    }
    else
    {
        *this = szModestBuffer;
    }
    va_end( wrk_args );
#endif

    return *this;
}

/************************************************************************/
/*                              FormatC()                               */
/************************************************************************/

/**
 * Format double in C locale.
 * 
 * The passed value is formatted using the C locale (period as decimal 
 * seperator) and appended to the target CPLString. 
 *
 * @param dfValue the value to format. 
 * @param pszFormat the sprintf() style format to use or omit for default.
 * Note that this format string should only include one substitution argument 
 * and it must be for a double (%f or %g). 
 *
 * @return a reference to the CPLString. 
 */

CPLString &CPLString::FormatC( double dfValue, const char *pszFormat )

{
    if( pszFormat == NULL )
        pszFormat = "%g";

    char szWork[512]; // presumably long enough for any number?

    sprintf( szWork, pszFormat, dfValue );
    CPLAssert( strlen(szWork) < sizeof(szWork) );
    
    if( strchr( szWork, ',' ) != NULL )
    {
        char *pszDelim = strchr( szWork, ',' );
        *pszDelim = '.';
    }

    *this += szWork;
    
    return *this;
}

/************************************************************************/
/*                                Trim()                                */
/************************************************************************/

/**
 * Trim white space.
 *
 * Trims white space off the let and right of the string.  White space
 * is any of a space, a tab, a newline ('\n') or a carriage control ('\r').
 *
 * @return a reference to the CPLString. 
 */

CPLString &CPLString::Trim()

{
    size_t iLeft, iRight;
    static const char szWhitespace[] = " \t\r\n";

    iLeft = find_first_not_of( szWhitespace );
    iRight = find_last_not_of( szWhitespace );

    if( iLeft == std::string::npos )
    {
        erase();
        return *this;
    }
    
    assign( substr( iLeft, iRight - iLeft + 1 ) );

    return *this;
}

/************************************************************************/
/*                               Recode()                               */
/************************************************************************/

CPLString &CPLString::Recode( const char *pszSrcEncoding,
                              const char *pszDstEncoding )

{
    if( pszSrcEncoding == NULL )
        pszSrcEncoding = CPL_ENC_UTF8;
    if( pszDstEncoding == NULL )
        pszDstEncoding = CPL_ENC_UTF8;

    if( strcmp(pszSrcEncoding,pszDstEncoding) == 0 )
        return *this;

    char *pszRecoded = CPLRecode( c_str(), 
                                  pszSrcEncoding,
                                  pszDstEncoding );

    if( pszRecoded == NULL )
        return *this;

    assign( pszRecoded );
    CPLFree( pszRecoded );

    return *this;
}

/************************************************************************/
/*                               ifind()                                */
/************************************************************************/

/** 
 * Case insensitive find() alternative.
 *
 * @param str substring to find.
 * @param pos offset in the string at which the search starts.
 * @return the position of substring in the string or std::string::npos if not found.
 * @since GDAL 1.9.0
 */

size_t CPLString::ifind( const std::string & str, size_t pos ) const

{
    return ifind( str.c_str(), pos );
}

/**
 * Case insensitive find() alternative.
 *
 * @param s substring to find.
 * @param nPos offset in the string at which the search starts.
 * @return the position of the substring in the string or std::string::npos if not found.
 * @since GDAL 1.9.0
 */

size_t CPLString::ifind( const char *s, size_t nPos ) const

{
    const char *pszHaystack = c_str();
    char chFirst = (char) ::tolower( s[0] );
    int nTargetLen = strlen(s);

    if( nPos > size() )
        nPos = size();

    pszHaystack += nPos;

    while( *pszHaystack != '\0' )
    {
        if( chFirst == ::tolower(*pszHaystack) )
        {
            if( EQUALN(pszHaystack,s,nTargetLen) )
                return nPos;
        }

        nPos++;
        pszHaystack++;
    }

    return std::string::npos;
}

/************************************************************************/
/*                              toupper()                               */
/************************************************************************/

/**
 * Convert to upper case in place.
 */

CPLString &CPLString::toupper()

{
    size_t i;

    for( i = 0; i < size(); i++ )
        (*this)[i] = (char) ::toupper( (*this)[i] );

    return *this;
}

/************************************************************************/
/*                              tolower()                               */
/************************************************************************/

/**
 * Convert to lower case in place.
 */

CPLString &CPLString::tolower()

{
    size_t i;

    for( i = 0; i < size(); i++ )
        (*this)[i] = (char) ::tolower( (*this)[i] );

    return *this;
}

/************************************************************************/
/*                         CPLURLGetValue()                             */
/************************************************************************/

/**
 * Return the value matching a key from a key=value pair in a URL.
 *
 * @param pszURL the URL.
 * @param pszKey the key to find.
 * @return the value of empty string if not found.
 * @since GDAL 1.9.0
 */
CPLString CPLURLGetValue(const char* pszURL, const char* pszKey)
{
    CPLString osKey(pszKey);
    osKey += "=";
    size_t nKeyPos = CPLString(pszURL).ifind(osKey.impl());
    if (nKeyPos != std::string::npos && nKeyPos > 0 &&
        (pszURL[nKeyPos-1] == '?' || pszURL[nKeyPos-1] == '&'))
    {
        CPLString osValue(pszURL + nKeyPos + strlen(osKey));
        const char* pszValue = osValue.c_str();
        const char* pszSep = strchr(pszValue, '&');
        if (pszSep)
        {
            osValue.resize(pszSep - pszValue);
        }
        return osValue;
    }
    return "";
}

/************************************************************************/
/*                          CPLURLAddKVP()                              */
/************************************************************************/

/**
 * Return a new URL with a new key=value pair.
 *
 * @param pszURL the URL.
 * @param pszKey the key to find.
 * @param pszValue the value of the key (may be NULL to unset an existing KVP).
 * @return the modified URL.
 * @since GDAL 1.9.0
 */
CPLString CPLURLAddKVP(const char* pszURL, const char* pszKey,
                       const char* pszValue)
{
    CPLString osURL(pszURL);
    if (strchr(osURL, '?') == NULL)
        osURL += "?";
    pszURL = osURL.c_str();

    CPLString osKey(pszKey);
    osKey += "=";
    size_t nKeyPos = osURL.ifind(osKey.impl());
    if (nKeyPos != std::string::npos && nKeyPos > 0 &&
        (pszURL[nKeyPos-1] == '?' || pszURL[nKeyPos-1] == '&'))
    {
        CPLString osNewURL(osURL);
        osNewURL.resize(nKeyPos);
        if (pszValue)
        {
            osNewURL += osKey;
            osNewURL += pszValue;
        }
        const char* pszNext = strchr(pszURL + nKeyPos, '&');
        if (pszNext)
        {
            if (osNewURL[osNewURL.size()-1] == '&'
                || osNewURL[osNewURL.size()-1] == '?' )
                osNewURL += pszNext + 1;
            else
                osNewURL += pszNext;
        }
        return osNewURL;
    }
    else
    {
        if (pszValue)
        {
            if (osURL[osURL.size()-1] != '&' && osURL[osURL.size()-1] != '?')
                osURL += '&';
            osURL += osKey;
            osURL += pszValue;
        }
        return osURL;
    }
}
